package fr.ippon.life;

import java.util.Collection;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public class Grid {
  private final Set<Cell> deadCells;
  private final Box box;

  Grid(Set<Cell> cells) {
    box = new Box(cells);
    this.deadCells = buildDeadCells(cells);
  }

  private Set<Cell> buildDeadCells(Set<Cell> cells) {
    return IntStream
      .range(box.getTopRow(), box.getBottomRow())
      .mapToObj(Integer::valueOf)
      .flatMap(row -> toDeadCell(cells, row))
      .collect(Collectors.toUnmodifiableSet());
  }

  private Stream<Cell> toDeadCell(Set<Cell> cells, int row) {
    return IntStream
      .range(box.getLeftColumn(), box.getRightColumn())
      .mapToObj(column -> new Cell(row, column))
      .filter(cell -> !cells.contains(cell));
  }

  public Collection<Cell> deadCells() {
    return deadCells;
  }

  private static class Box {
    private final Point topLeft;
    private final Point bottomRight;

    public Box(Set<Cell> cells) {
      topLeft = Point.topLeft(cells);
      bottomRight = Point.bottomRight(cells);
    }

    public int getTopRow() {
      return topLeft.getRow();
    }

    public int getBottomRow() {
      return bottomRight.getRow() + 1;
    }

    public int getLeftColumn() {
      return topLeft.getColumn();
    }

    public int getRightColumn() {
      return bottomRight.getColumn() + 1;
    }
  }

  private static class Point {
    private final int row;
    private final int column;

    public Point(int row, int column) {
      this.row = row;
      this.column = column;
    }

    private static Point topLeft(Set<Cell> cells) {
      int row = cells.stream().mapToInt(Cell::getRow).min().orElse(0);

      int column = cells.stream().mapToInt(Cell::getColumn).min().orElse(0);

      return new Point(row - 1, column - 1);
    }

    private static Point bottomRight(Set<Cell> cells) {
      int row = cells.stream().mapToInt(Cell::getRow).max().orElse(0);

      int column = cells.stream().mapToInt(Cell::getColumn).max().orElse(0);

      return new Point(row + 1, column + 1);
    }

    public int getRow() {
      return row;
    }

    public int getColumn() {
      return column;
    }
  }
}
